unit waveio;

interface

uses
  Windows, Classes, MMSystem, SysUtils, msacm, Dialogs;

const
  MPEGLAYER3 = 85;
  
type
  EWaveIOError = Exception;

  TWaveStream = class(TObject)
  private
    FFormat: PWaveFormatEx;
    FDataOffset, FSize: Longint;
    mm: hmmIO;
    pck, pckRIFF: TMMCKINFO;

    function GetPosition: Longint;
    procedure SetPosition(Pos: Longint);
    procedure CheckMMIOWave;

  public
    constructor Create( mmIO: hmmIO; WriteFormat: PWaveFormatEx );
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; virtual;
    function Seek(Offset: Longint; Origin: Word): Longint; virtual;
    function Write(var Buffer; Count: Longint): Longint; virtual;

    property Format: PWaveFormatEx read FFormat;
    property Position: Longint read GetPosition write SetPosition;
    property Size: Longint read FSize;
  end;

  TMemoryWaveStream = class(TWaveStream)
  private
    mmIO: hmmIO;

  public
    constructor Create( Memory: Pointer; Size: Longint;
      WriteFormat: PWaveFormatEx ); virtual;
    destructor Destroy; override;
  end;

  TFileWaveStream = class(TWaveStream)
  private
    mmIO: hmmIO;

  public
    constructor Create( FileName: String; WriteFormat: PWaveFormatEx ); virtual;
    destructor Destroy; override;
  end;

  TPCMWaveReader = class(TObject)
  private
    FPosition: Longint;
    FStream: TWaveStream;
    FSize: Longint;
    FBufferLength: Longint;
    acmStream: Integer;
    ash: TACMStreamHeader;
    dstwfx: PWaveFormatEx;

    RawBuffer: Pointer;
    RawBufferByteSize: Longint;
    RawBufferSampleSize: Longint;

    PCMBuffer: Pointer;
    PCMBufferByteSize: Longint;     // size of total buffer in bytes
    PCMBufferSampleSize: Longint;   // size of converted samples in buffer

    PCMSamplesPerSample, PCMBufferSamplePos: Longint;

    procedure AllocBuffers;
    procedure DestroyBuffers;
    procedure ReadSamples(Index: Longint);
    procedure SetBufferLength( Value: Longint );
    procedure SetPosition( Value: Longint );

  public
    constructor Create( Stream: TWaveStream );
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint;

    property Format: PWaveFormatEx read dstwfx;
    property BufferLength: Longint read FBufferLength write SetBufferLength;
    property Position: Longint read FPosition write SetPosition;
    property Size: Longint read FSize;
    property Stream: TWaveStream read FStream;
  end;


implementation


function mmioFourCC( Chr1: Char; Chr2: Char; Chr3: Char; Chr4: Char ): DWord;
begin
  Result := Integer(Chr1) + (Integer(Chr2) shl 8) + (Integer(Chr3) shl 16)
    + (Integer(Chr4) shl 24);
end;


constructor TWaveStream.Create( mmIO: hmmIO; WriteFormat: PWaveFormatEx );
begin
  inherited Create;

  mm := mmIO;

  if (WriteFormat <> nil) then
  begin
    // Create the output file RIFF chunk of form type 'WAVE'.
    pckRIFF.fccType := mmioFOURCC('W', 'A', 'V', 'E');
    pckRIFF.cksize := 0;
    if (mmioCreateChunk(mm, @pckRIFF, MMIO_CREATERIFF) <> 0) then
      raise EWaveIOError.Create('Error creating chunk.');

    // We are now descended into the 'RIFF' chunk we just created.
    // * Now create the 'fmt ' chunk. Since we know the size of this chunk,
    // specify it in the MMCKINFO structure so MMIO doesn't have to seek
    // back and set the chunk size after ascending from the chunk.
    //
    pck.ckid := mmioFOURCC('f', 'm', 't', ' ');
    pck.cksize := sizeof(WriteFormat^) + WriteFormat^.cbSize;
    if (mmioCreateChunk(mm, @pck, 0) <> 0) then
      raise EWaveIOError.Create('Error creating chunk.');

    // Write the variable length size.
    if (mmioWrite(mm, Pointer(WriteFormat), sizeof(WriteFormat^) + WriteFormat^.cbSize)
      <> sizeof(WriteFormat^) + WriteFormat^.cbSize) then
      raise EWaveIOError.Create('Error writing wave format.');

    GetMem( FFormat, sizeof(FFormat^) );
    CopyMemory( FFormat, WriteFormat, sizeof(FFormat^) );

    // Ascend out of the 'fmt ' chunk, back into the 'RIFF' chunk.
    if (mmioAscend(mm, @pck, 0) <> 0) then
      raise EWaveIOError.Create('Cannot ascend chunk.');

    // We are now descended into the 'RIFF' chunk we just created.
    // * Now create the 'data' chunk.
    pck.ckid := mmioFOURCC('d', 'a', 't', 'a');
    pck.cksize := 0;
    if (mmioCreateChunk(mm, @pck, 0) <> 0) then
      raise EWaveIOError.Create('Error creating chunk.');
  end
  else
  begin
    CheckMMIOWave;

    // put in the beggining of the file
    mmioSeek( mm, FDataOffset, SEEK_SET );
  end;
end;


destructor TWaveStream.Destroy;
begin
  if (FFormat <> nil) then
    FreeMem( FFormat );

  mmioSeek( mm, 0, SEEK_END );

  if (mmioAscend(mm, @pck, 0) <> 0) then
    raise EWaveIOError.Create('Cannot ascend chunk.');

  if (mmioAscend(mm, @pckRIFF, 0) <> 0) then
    raise EWaveIOError.Create('Cannot ascend chunk.');

  if (mmioFlush(mm, 0) <> 0) then
    raise EWaveIOError.Create('Cannot flush.');

  inherited Destroy;
end;


function TWaveStream.GetPosition: Longint;
begin
  Result := Seek(0, SEEK_CUR);
end;

procedure TWaveStream.SetPosition(Pos: Longint);
begin
  Seek(Pos, SEEK_SET);
end;

function TWaveStream.Read(var Buffer; Count: Longint): Longint;
var
  p: Pointer;
begin
  p := @Buffer;

  Result := mmioRead( mm, p, Count * FFormat^.nBlockAlign );
  if (Result = -1) then
    raise EWaveIOError.Create('Error reading from file.')
  else
    Result := Result div FFormat^.nBlockAlign;
end;

function TWaveStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  if (Origin = SEEK_SET) then
    Result := (mmioSeek( mm, Offset * FFormat^.nBlockAlign + FDataOffset,
      Origin ) - FDataOffset) div FFormat^.nBlockAlign
  else
    Result := (mmioSeek( mm, Offset * FFormat^.nBlockAlign,
      Origin ) - FDataOffset) div FFormat^.nBlockAlign;

  if (Result < 0) then
    raise EWaveIOError.Create('Seeking to an invalid location in the file.');
end;

function TWaveStream.Write(var Buffer; Count: Longint): Longint;
begin
  Result := mmioWrite( mm, @Buffer, Count * FFormat^.nBlockAlign );
  if (Result = -1) then
    raise EWaveIOError.Create('Error writing to file.')
  else
    Result := Result div FFormat^.nBlockAlign;
end;

procedure TWaveStream.CheckMMIOWave;
var
  mmIOInfo: TMMIOInfo;
  FormatTmp: TWaveFormatEx;
  ExtraAlloc: Word;
begin
  if (mmioDescend(mm, @pckRIFF, nil, 0) <> 0) then
    raise EWaveIOError.Create( 'Invalid multimedia file!' );

  if (pckRIFF.ckid <> FOURCC_RIFF)
  or (pckRIFF.fccType <> mmioFOURCC('W', 'A', 'V', 'E')) then
    raise EWaveIOError.Create( 'Not a wave file!' );

  // Search the input file for for the 'fmt ' chunk.     */
  pck.ckid := mmioFOURCC('f', 'm', 't', ' ');
  if (mmioDescend(mm, @pck, @pckRIFF, MMIO_FINDCHUNK) <> 0) then
    raise EWaveIOError.Create( 'Couldn''t find ''fmt'' chunk!' );

  // Expect the 'fmt' chunk to be at least as large as <PCMWAVEFORMAT>;
  // if there are extra parameters at the end, we'll ignore them */
  if (pck.cksize < 16) then
    raise EWaveIOError.Create( 'Abnormal ''fmt'' size!' );

  // Read the 'fmt ' chunk into <pcmWaveFormat>.*/
  if (mmioRead(mm, PChar(@FormatTmp), 16) <> 16) then
    raise EWaveIOError.Create( 'Error reading ''fmt'' chunk!' );

  // Ok, allocate the waveformatex, but if its not pcm
  // format, read the next word, and thats how many extra
  // bytes to allocate.
  if (FormatTmp.wFormatTag = WAVE_FORMAT_PCM) then
    ExtraAlloc := 0
  else
    // Read in length of extra bytes.
    if (mmioRead(mm, @ExtraAlloc, sizeof(ExtraAlloc)) <>
      sizeof(ExtraAlloc)) then
      raise EWaveIOError.Create( 'Error reading ''waveformatex'' length!' );

  GetMem( FFormat, sizeof(FFormat^)+ExtraAlloc );
  CopyMemory( FFormat, @FormatTmp, sizeof(FFormat^) );
  FFormat^.cbSize := ExtraAlloc;
  if (ExtraAlloc <> 0) then
    if (mmioRead(mm, PChar(FFormat) + sizeof(FFormat^), ExtraAlloc) <>
      ExtraAlloc) then
      raise EWaveIOError.Create( 'Error reading ''waveformatex''!' );

  if (FFormat^.wFormatTag = MPEGLAYER3) then
    raise EWaveIOError.Create( 'MPEG Layer-3 compression is not supported.' );

  // Ascend the input file out of the 'fmt ' chunk. */
  if (mmioAscend(mm, @pck, 0) <> 0) then
    raise EWaveIOError.Create( 'Error ascending from ''fmt'' chunk!' );

  // Do a nice little seek...
  if (mmioSeek(mm, pckRIFF.dwDataOffset + sizeof(FOURCC), SEEK_SET) = -1) then
    raise EWaveIOError.Create( 'Error seeking data!' );

  //      Search the input file for the 'data' chunk.
  pck.ckid := mmioFOURCC('d', 'a', 't', 'a');
  mmioDescend(mm, @pck, @pckRIFF, MMIO_FINDCHUNK);

  if (mmioGetInfo(mm, @mmioInfo, 0) <> 0) then
    raise EWaveIOError.Create('Error getting info!');

  FDataOffset := pck.dwDataOffset;
  FSize := pck.cksize div FFormat^.nBlockAlign;
end;


//***************************************************************
//        TMemoryWaveStream
//***************************************************************

constructor TMemoryWaveStream.Create( Memory: Pointer; Size: Longint;
  WriteFormat: PWaveFormatEx );
var
  info: TMMIOINFO;
begin
  ZeroMemory( @info, sizeof(info) );
  with info do
  begin
    pchBuffer := Memory;
    fccIOProc := FOURCC_MEM;
    cchBuffer := Size;
  end;

  // Initialization...
  mmIO := mmioOpen( nil,  @info, MMIO_READ );
  if (mmIO = 0) then
    raise EWaveIOError.Create('Can''t open memory stream.');

  inherited Create( mmIO, WriteFormat );
end;

destructor TMemoryWaveStream.Destroy;
begin
  inherited Destroy;

  if (mmIO <> 0) then mmioClose( mmIO, 0);
end;


//***************************************************************
//        TFileWaveStream
//***************************************************************

constructor TFileWaveStream.Create( FileName: String;
  WriteFormat: PWaveFormatEx );
begin
  // Initialization...
  if (WriteFormat = nil) then
    mmIO := mmioOpen( Pointer(FileName),  nil, MMIO_READ or MMIO_ALLOCBUF )
  else
    mmIO := mmioOpen( Pointer(FileName),  nil, MMIO_CREATE or MMIO_READWRITE
      or MMIO_ALLOCBUF );

  if (mmIO = 0) then
    raise EWaveIOError.Create('Can''t open file stream.');

  inherited Create( mmIO, WriteFormat );
end;

destructor TFileWaveStream.Destroy;
begin
  inherited Destroy;

  if (mmIO <> 0) then mmioClose( mmIO, 0);
end;


//***************************************************************
//        TPCMWaveReader
//***************************************************************

constructor TPCMWaveReader.Create( Stream: TWaveStream );
begin
  inherited Create;

  FStream := Stream;
  GetMem( dstwfx, sizeof(dstwfx^) );

//  FBufferLength := 4096;
  FBufferLength := 20000;

  if (FStream.Format^.wFormatTag <> WAVE_FORMAT_PCM) then
  begin
    //
    // prepare acm stream converter
    //
    ZeroMemory( dstwfx, sizeof(dstwfx^) );
    dstwfx^.wFormatTag := WAVE_FORMAT_PCM;
    if (acmFormatSuggest( 0, FStream.FFormat, dstwfx, sizeof(dstwfx^),
      ACM_FORMATSUGGESTF_WFORMATTAG ) <> 0) then
      raise EWaveIOError.Create('Can''t suggest pcm format.');

    if (acmStreamOpen( PHACMSTREAM(@acmStream), 0, FStream.Format^, dstwfx^,
      nil, 0, 0, 0 ) <> 0) then
      raise EWaveIOError.Create('Can''t open acm stream.');

    AllocBuffers;

    // prepare buffers
    ZeroMemory( @ash, sizeof(ash) );
    with ash do
    begin
      cbStruct := sizeof(ash);
      pbSrc := Pointer(RawBuffer);
      cbSrcLength := RawBufferByteSize;
      dwSrcUser := cbSrcLength;
      pbDst := Pointer(PCMBuffer);
      cbDstLength := PCMBufferByteSize;
      dwDstUser := cbDstLength;
    end;

    if (acmStreamPrepareHeader( acmStream, ash, 0 ) <> 0) then
      raise EWaveIOError.Create('Can''t prepare headers.');
  end
  else
    dstwfx^ := FStream.Format^;

  //
  if (FStream.Format^.wFormatTag <> WAVE_FORMAT_PCM) then
    PCMSamplesperSample := (PCMBufferByteSize * FStream.Format^.nBlockAlign)
      div (RawBufferByteSize * dstwfx^.nBlockAlign)
  else
    PCMSamplesPerSample := 1;
  FSize := PCMSamplesPerSample * FStream.Size;
  PCMBufferSampleSize := 0;
end;


destructor TPCMWaveReader.Destroy;
begin
  if (acmStream <> 0) then
  begin
    ash.cbSrcLength := RawBufferByteSize;
    acmStreamUnprepareHeader( acmStream, ash, 0 );
    acmStreamClose( acmStream, 0 );
  end;

  if (dstwfx <> nil) then
    FreeMem(dstwfx);

  DestroyBuffers;

  inherited Destroy;
end;


procedure TPCMWaveReader.AllocBuffers;
var
   ss, dd: DWord;
begin
  DestroyBuffers;

  // calc space needed for holding the decompression of one sample
  if (acmStreamSize( acmStream, FStream.Format^.nBlockAlign, dd,
    ACM_STREAMSIZEF_SOURCE ) <> 0) then
    raise EWaveIOError.Create('Can''t recommend an acm stream size.');

  // calc minimum size block of the source
  if (acmStreamSize( acmStream, dd, ss, ACM_STREAMSIZEF_DESTINATION ) <> 0) then
    raise EWaveIOError.Create('Can''t recommend an acm stream size.');

  // alloc source buffer (raw)
  RawBufferSampleSize := FBufferLength div dd;
  if (RawBufferSampleSize = 0) then RawBufferSampleSize := 1;
  RawBufferByteSize := RawBufferSampleSize * ss;
  GetMem( RawBuffer, RawBufferByteSize );

  // Alloc destination buffer (decompressed)
  if (acmStreamSize( acmStream, DWord(RawBufferByteSize),
    DWord(PCMBufferByteSize), ACM_STREAMSIZEF_SOURCE ) <> 0) then
    raise EWaveIOError.Create('Can''t recommend an acm stream size.');
  GetMem( PCMBuffer, PCMBufferByteSize );
end;


procedure TPCMWaveReader.DestroyBuffers;
begin
  PCMBufferSampleSize := 0;

  if (PCMBuffer <> nil) then
    FreeMem(PCMBuffer);
  PCMBuffer := nil;

  if (RawBuffer <> nil) then
    FreeMem(RawBuffer);
  RawBuffer := nil;
end;

procedure TPCMWaveReader.SetBufferLength( Value: Longint );
begin
  if (Value <> FBufferLength) and (Value >= 1024) then
  begin
    FBufferLength := Value;
    if (PCMBuffer <> nil) then
      AllocBuffers;
  end;
end;

procedure TPCMWaveReader.SetPosition( Value: Longint );
begin
  if (Value <> FPosition) then
    if (Value >= 0) and (Value < FSize) then
      FPosition := Value
    else
      raise EWaveIOError.Create('Position out of bounds.');
end;

function TPCMWaveReader.Read(var Buffer; Count: Longint): Longint;
var
  pos, len: Longint;
  posi, posf: Longint;
begin
  if (Count < 1) then
  begin
    Result := 0;
    exit;
  end;

  if (FStream.Format^.wFormatTag = WAVE_FORMAT_PCM) then
  begin
    FStream.Position := FPosition;
    Result := FStream.Read( Buffer, Count );
    FPosition := FPosition + Result;
  end
  else
  begin
    // limit to wave size
    if (Count + FPosition >= FSize) then
      Count := FSize - FPosition;

    // use current buffer data if possible
    if (FPosition >= PCMBufferSamplePos)
      and (FPosition < PCMBufferSamplePos + PCMBufferSampleSize) then
    begin
      len := PCMBufferSamplePos + PCMBufferSampleSize - FPosition;
      if (len > Count) then len := Count;

      CopyMemory( PChar(@Buffer),
          PChar(PCMBuffer) + (FPosition-PCMBufferSamplePos)*dstwfx^.nBlockAlign,
          len*dstwfx^.nBlockAlign );

      pos := len;
    end
    else
      pos := 0;

    // put next data
    while (pos < Count) do
    begin
      ReadSamples( (FPosition+pos) div PCMSamplesPerSample );

      // calc. range of current pcm buffer that can be used to fill request
      posi := FPosition+pos;
      if (FPosition+pos < PCMBufferSamplePos) then
        raise EWaveIOError.Create('Internal error: Position smaller than PCMBufferSamplePos');

      if (FPosition + Count > PCMBufferSamplePos + PCMBufferSampleSize) then
        posf := PCMBufferSamplePos + PCMBufferSampleSize
      else posf := FPosition + Count;

      len := posf-posi;

      // put pcm buffer data into target
      CopyMemory( PChar(@Buffer) + (posi-FPosition)*dstwfx^.nBlockAlign,
        PChar(PCMBuffer) + (posi-PCMBufferSamplePos)*dstwfx^.nBlockAlign,
        len*dstwfx^.nBlockAlign );

      pos := pos + len;
    end;

    FPosition := FPosition + Count;
    Result := Count;
  end;
end;


procedure TPCMWaveReader.ReadSamples(Index: Longint);
begin
  FStream.Position := Index;

  ash.cbSrcLength := FStream.Read( RawBuffer^, RawBufferSampleSize )
    * FStream.Format^.nBlockAlign;
  if (acmStreamConvert( acmStream, ash,
    ACM_STREAMCONVERTF_BLOCKALIGN ) <> 0) then
    raise EWaveIOError.Create('Unable to convert sample.');

  PCMBufferSampleSize := ash.cbDstLengthUsed div dstwfx^.nBlockAlign;
  PCMBufferSamplePos := Index * PCMSamplesPerSample;
end;


end.
